package java.lang.invoke;

import java.io.InputStream;
import java.net.URLConnection;
import java.io.IOException;
import sun.reflect.CallerSensitive;
import sun.reflect.Reflection;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.Collections;
import java.util.ArrayList;
import sun.invoke.empty.Empty;
import java.lang.reflect.Array;
import java.util.Arrays;
import sun.invoke.util.ValueConversions;
import sun.invoke.util.Wrapper;
import sun.invoke.util.VerifyType;
import java.util.function.Function;

abstract class MethodHandleImpl
{
    private static final int MAX_ARITY;
    private static final Function<MethodHandle, LambdaForm> PRODUCE_BLOCK_INLINING_FORM;
    private static final Function<MethodHandle, LambdaForm> PRODUCE_REINVOKER_FORM;
    static MethodHandle[] FAKE_METHOD_HANDLE_INVOKE;
    private static final Object[] NO_ARGS_ARRAY;
    private static final int FILL_ARRAYS_COUNT = 11;
    private static final int LEFT_ARGS = 10;
    private static final MethodHandle[] FILL_ARRAY_TO_RIGHT;
    private static final ClassValue<MethodHandle[]> TYPED_COLLECTORS;
    static final int MAX_JVM_ARITY = 255;
    
    MethodHandleImpl() {
    }
    
    static void initStatics() {
        MemberName.Factory.INSTANCE.getClass();
    }
    
    static MethodHandle makeArrayElementAccessor(Class<?> clazz, boolean b) {
        if (clazz == Object[].class) {
            return b ? MethodHandleImpl.ArrayAccessor.OBJECT_ARRAY_SETTER : MethodHandleImpl.ArrayAccessor.OBJECT_ARRAY_GETTER;
        }
        if (!clazz.isArray()) {
            throw MethodHandleStatics.newIllegalArgumentException("not an array: " + clazz);
        }
        MethodHandle[] array = (MethodHandle[])MethodHandleImpl.ArrayAccessor.TYPED_ACCESSORS.get(clazz);
        int n = b;
        MethodHandle methodHandle = array[n];
        if (methodHandle != null) {
            return methodHandle;
        }
        MethodHandle methodHandle2 = MethodHandleImpl.ArrayAccessor.getAccessor(clazz, b);
        MethodType correctType = MethodHandleImpl.ArrayAccessor.correctType(clazz, b);
        if (methodHandle2.type() != correctType) {
            assert methodHandle2.type().parameterType(0) == Object[].class;
            assert (b ? methodHandle2.type().parameterType(2) : methodHandle2.type().returnType()) == Object.class;
            assert correctType.parameterType(0).getComponentType() == correctType.returnType();
            methodHandle2 = methodHandle2.viewAsType(correctType, false);
        }
        MethodHandle intrinsic = makeIntrinsic(methodHandle2, b ? MethodHandleImpl.Intrinsic.ARRAY_STORE : MethodHandleImpl.Intrinsic.ARRAY_LOAD);
        synchronized (array) {
            if (array[n] == null) {
                array[n] = intrinsic;
            }
            else {
                intrinsic = array[n];
            }
        }
        return intrinsic;
    }
    
    static MethodHandle makePairwiseConvert(MethodHandle methodHandle, MethodType methodType, boolean b, boolean b2) {
        if (methodType == methodHandle.type()) {
            return methodHandle;
        }
        return makePairwiseConvertByEditor(methodHandle, methodType, b, b2);
    }
    
    private static int countNonNull(Object[] array) {
        int n = 0;
        for (int length = array.length, i = 0; i < length; ++i) {
            if (array[i] != null) {
                ++n;
            }
        }
        return n;
    }
    
    static MethodHandle makePairwiseConvertByEditor(MethodHandle methodHandle, MethodType methodType, boolean b, boolean b2) {
        Object[] computeValueConversions = computeValueConversions(methodType, methodHandle.type(), b, b2);
        int countNonNull = countNonNull(computeValueConversions);
        if (countNonNull == 0) {
            return methodHandle.viewAsType(methodType, b);
        }
        MethodType basicType = methodType.basicType();
        MethodType methodType2 = methodHandle.type().basicType();
        BoundMethodHandle boundMethodHandle = methodHandle.rebind();
        for (int i = 0; i < computeValueConversions.length - 1; ++i) {
            Object o = computeValueConversions[i];
            if (o != null) {
                MethodHandle bindTo;
                if ((o instanceof Class)) {
                    bindTo = MethodHandleImpl.Lazy.MH_castReference.bindTo(o);
                }
                else {
                    bindTo = (MethodHandle)o;
                }
                Class<Object> parameterType = basicType.parameterType(i);
                if (--countNonNull == 0) {
                    methodType2 = methodType;
                }
                else {
                    methodType2 = methodType2.changeParameterType(i, parameterType);
                }
                boundMethodHandle = boundMethodHandle.copyWithExtendL(methodType2, boundMethodHandle.editor().filterArgumentForm(1 + i, LambdaForm.BasicType.basicType(parameterType)), bindTo).rebind();
            }
        }
        Object o2 = computeValueConversions[computeValueConversions.length - 1];
        if (o2 != null) {
            MethodHandle bindTo2;
            if ((o2 instanceof Class)) {
                if (o2 == Void.TYPE) {
                    bindTo2 = null;
                }
                else {
                    bindTo2 = MethodHandleImpl.Lazy.MH_castReference.bindTo(o2);
                }
            }
            else {
                bindTo2 = (MethodHandle)o2;
            }
            Class<Object> returnType = basicType.returnType();
            assert --countNonNull == 0;
            if (bindTo2 != null) {
                BoundMethodHandle rebind = boundMethodHandle.rebind();
                boundMethodHandle = rebind.copyWithExtendL(methodType, rebind.editor().filterReturnForm(LambdaForm.BasicType.basicType(returnType), false), bindTo2);
            }
            else {
                boundMethodHandle = boundMethodHandle.copyWith(methodType, boundMethodHandle.editor().filterReturnForm(LambdaForm.BasicType.basicType(returnType), true));
            }
        }
        assert countNonNull == 0;
        assert !(!boundMethodHandle.type().equals(methodType));
        return boundMethodHandle;
    }
    
    static MethodHandle makePairwiseConvertIndirect(MethodHandle methodHandle, MethodType methodType, boolean b, boolean b2) {
        assert methodHandle.type().parameterCount() == methodType.parameterCount();
        Object[] computeValueConversions = computeValueConversions(methodType, methodHandle.type(), b, b2);
        int parameterCount = methodType.parameterCount();
        int countNonNull = countNonNull(computeValueConversions);
        int n = (computeValueConversions[parameterCount] != null) ? 1 : 0;
        int n2 = methodType.returnType() == Void.TYPE;
        if (n != 0 && n2) {
            --countNonNull;
            n = 0;
        }
        int n3 = 1 + parameterCount;
        int n4 = n3 + countNonNull + 1;
        int n5 = (n == 0) ? -1 : (n4 - 1);
        int n6 = ((n == 0) ? n4 : n5) - 1;
        int n7 = n2 ? -1 : (n4 - 1);
        MethodType invokerType = methodType.basicType().invokerType();
        LambdaForm.Name[] arguments = LambdaForm.arguments(n4 - n3, invokerType);
        Object[] array = new Object[0 + parameterCount];
        for (int i = 0; i < parameterCount; ++i) {
            Object o = computeValueConversions[i];
            if (o == null) {
                array[0 + i] = arguments[1 + i];
            }
            else {
                Object o2;
                if ((o instanceof Class)) {
                    o2 = new LambdaForm.Name(MethodHandleImpl.Lazy.MH_castReference, new Object[] { (Class)o, arguments[1 + i] });
                }
                else {
                    o2 = new LambdaForm.Name((MethodHandle)o, new Object[] { arguments[1 + i] });
                }
                assert arguments[n3] == null;
                arguments[n3++] = o2;
                assert array[0 + i] == null;
                array[0 + i] = o2;
            }
        }
        assert n3 == n6;
        arguments[n6] = new LambdaForm.Name(methodHandle, array);
        Object o3 = computeValueConversions[parameterCount];
        if (n == 0) {
            assert n6 == arguments.length - 1;
        }
        else {
            Object o4;
            if (o3 == Void.TYPE) {
                o4 = new LambdaForm.Name(LambdaForm.constantZero(LambdaForm.BasicType.basicType(methodType.returnType())), new Object[0]);
            }
            else if ((o3 instanceof Class)) {
                o4 = new LambdaForm.Name(MethodHandleImpl.Lazy.MH_castReference, new Object[] { (Class<Void>)o3, arguments[n6] });
            }
            else {
                MethodHandle methodHandle2 = (MethodHandle)o3;
                if (methodHandle2.type().parameterCount() == 0) {
                    o4 = new LambdaForm.Name(methodHandle2, new Object[0]);
                }
                else {
                    o4 = new LambdaForm.Name(methodHandle2, new Object[] { arguments[n6] });
                }
            }
            assert arguments[n5] == null;
            arguments[n5] = o4;
            assert n5 == arguments.length - 1;
        }
        return SimpleMethodHandle.make(methodType, new LambdaForm("convert", invokerType.parameterCount(), arguments, n7));
    }
    
    @ForceInline
    static <T, U> T castReference(Class<? extends T> clazz, U u) {
        if (u != null && !clazz.isInstance(u)) {
            throw newClassCastException(clazz, u);
        }
        return u;
    }
    
    private static ClassCastException newClassCastException(Class<?> clazz, Object o) {
        return new ClassCastException("Cannot cast " + o.getClass().getName() + " to " + clazz.getName());
    }
    
    static Object[] computeValueConversions(MethodType methodType, MethodType methodType2, boolean keepInterfaces, boolean b) {
        int parameterCount = methodType.parameterCount();
        Object[] array = new Object[parameterCount + 1];
        for (int i = 0; i <= parameterCount; ++i) {
            int n = i == parameterCount;
            Class<Object> src = n ? methodType2.returnType() : methodType.parameterType(i);
            Class<Object> dst = n ? methodType.returnType() : methodType2.parameterType(i);
            if (!VerifyType.isNullConversion(src, dst, keepInterfaces)) {
                array[i] = valueConversion(src, dst, keepInterfaces, b);
            }
        }
        return array;
    }
    
    static MethodHandle makePairwiseConvert(MethodHandle methodHandle, MethodType methodType, boolean b) {
        return makePairwiseConvert(methodHandle, methodType, b, false);
    }
    
    static Object valueConversion(Class<?> src, Class<?> clazz, boolean keepInterfaces, boolean b) {
        assert !VerifyType.isNullConversion(src, clazz, keepInterfaces);
        if (clazz == Void.TYPE) {
            return clazz;
        }
        MethodHandle methodHandle;
        if (src.isPrimitive()) {
            if (src == Void.TYPE) {
                return Void.TYPE;
            }
            if (clazz.isPrimitive()) {
                methodHandle = ValueConversions.convertPrimitive(src, clazz);
            }
            else {
                Wrapper forPrimitiveType2 = Wrapper.forPrimitiveType(src);
                methodHandle = ValueConversions.boxExact(forPrimitiveType2);
                assert methodHandle.type().parameterType(0) == forPrimitiveType2.primitiveType();
                assert methodHandle.type().returnType() == forPrimitiveType2.wrapperType();
                if (!VerifyType.isNullConversion(forPrimitiveType2.wrapperType(), clazz, keepInterfaces)) {
                    MethodType methodType = MethodType.methodType(clazz, src);
                    if (keepInterfaces) {
                        methodHandle = methodHandle.asType(methodType);
                    }
                    else {
                        methodHandle = makePairwiseConvert(methodHandle, methodType, false);
                    }
                }
            }
        }
        else if (clazz.isPrimitive()) {
            Wrapper forPrimitiveType = Wrapper.forPrimitiveType(clazz);
            if (b || src == forPrimitiveType.wrapperType()) {
                methodHandle = ValueConversions.unboxExact(forPrimitiveType, keepInterfaces);
            }
            else {
                methodHandle = (keepInterfaces ? ValueConversions.unboxWiden(forPrimitiveType) : ValueConversions.unboxCast(forPrimitiveType));
            }
        }
        else {
            return clazz;
        }
        assert methodHandle.type().parameterCount() <= 1 : "pc" + Arrays.asList(new Object[] { src.getSimpleName(), clazz.getSimpleName(), methodHandle });
        return methodHandle;
    }
    
    static MethodHandle makeVarargsCollector(MethodHandle methodHandle, Class<?> clazz) {
        MethodType type = methodHandle.type();
        int n = type.parameterCount() - 1;
        if (type.parameterType(n) != clazz) {
            methodHandle = methodHandle.asType(type.changeParameterType(n, clazz));
        }
        methodHandle = methodHandle.asFixedArity();
        return new MethodHandleImpl.AsVarargsCollector(methodHandle, clazz);
    }
    
    static MethodHandle makeSpreadArguments(MethodHandle type, Class<?> clazz, int n, int i) {
        MethodType methodType = type.type();
        Class<Object> spreadArgElementType;
        for (int j = 0; j < i; methodType = methodType.changeParameterType(n + j, spreadArgElementType), ++j) {
            spreadArgElementType = VerifyType.spreadArgElementType(clazz, j);
            if (spreadArgElementType == null) {
                spreadArgElementType = Object.class;
            }
        }
        type = type.asType(methodType);
        MethodType replaceParameterTypes = methodType.replaceParameterTypes(n, n + i, new Class[] { clazz });
        MethodType invokerType = replaceParameterTypes.invokerType();
        LambdaForm.Name[] arguments = LambdaForm.arguments(i + 2, invokerType);
        int parameterCount = invokerType.parameterCount();
        int[] array = new int[methodType.parameterCount()];
        for (int k = 0, n2 = 1; k < methodType.parameterCount() + 1; ++k, ++n2) {
            invokerType.parameterType(k);
            if (k == n) {
                MethodHandle arrayElementGetter = MethodHandles.arrayElementGetter(clazz);
                LambdaForm.Name name = arguments[n2];
                arguments[parameterCount++] = new LambdaForm.Name(MethodHandleImpl.Lazy.NF_checkSpreadArgument, new Object[] { name, Integer.valueOf(i) });
                for (int l = 0; l < i; ++l) {
                    array[k] = parameterCount;
                    arguments[parameterCount++] = new LambdaForm.Name(arrayElementGetter, new Object[] { name, Integer.valueOf(l) });
                    ++k;
                }
            }
            else if (k < array.length) {
                array[k] = n2;
            }
        }
        assert parameterCount == arguments.length - 1;
        LambdaForm.Name[] array2 = new LambdaForm.Name[methodType.parameterCount()];
        for (int n3 = 0; n3 < methodType.parameterCount(); ++n3) {
            array2[n3] = arguments[array[n3]];
        }
        arguments[arguments.length - 1] = new LambdaForm.Name(type, (Object[])array2);
        return SimpleMethodHandle.make(replaceParameterTypes, new LambdaForm("spread", invokerType.parameterCount(), arguments));
    }
    
    static void checkSpreadArgument(Object o, int i) {
        if (o == null) {
            if (i == 0) {
                return;
            }
        }
        else if ((o instanceof Object[])) {
            if (((Object[])(Object[])o).length == i) {
                return;
            }
        }
        else if (Array.getLength(o) == i) {
            return;
        }
        throw MethodHandleStatics.newIllegalArgumentException("array is not of length " + i);
    }
    
    static MethodHandle makeCollectArguments(MethodHandle methodHandle, MethodHandle methodHandle2, int n, boolean b) {
        MethodType type = methodHandle.type();
        MethodType type2 = methodHandle2.type();
        int parameterCount = type2.parameterCount();
        Class<Object> returnType = type2.returnType();
        MethodType methodType = type.dropParameterTypes(n, n + ((returnType == Void.TYPE) ? 0 : 1));
        if (!b) {
            methodType = methodType.insertParameterTypes(n, type2.parameterList());
        }
        MethodType invokerType = methodType.invokerType();
        LambdaForm.Name[] arguments = LambdaForm.arguments(2, invokerType);
        int n2 = arguments.length - 2;
        int n3 = arguments.length - 1;
        arguments[n2] = new LambdaForm.Name(methodHandle2, (Object[])(LambdaForm.Name[])Arrays.copyOfRange(arguments, 1 + n, 1 + n + parameterCount));
        LambdaForm.Name[] array = new LambdaForm.Name[type.parameterCount()];
        int n4 = 1;
        int n5 = 0;
        System.arraycopy(arguments, n4, array, n5, n);
        int n6 = n4 + n;
        int n7 = n5 + n;
        if (returnType != Void.TYPE) {
            array[n7++] = arguments[n2];
        }
        if (b) {
            System.arraycopy(arguments, n6, array, n7, parameterCount);
            n7 += parameterCount;
        }
        int n8 = n6 + parameterCount;
        int n9 = array.length - n7;
        System.arraycopy(arguments, n8, array, n7, n9);
        assert n8 + n9 == n2;
        arguments[n3] = new LambdaForm.Name(methodHandle, (Object[])array);
        return SimpleMethodHandle.make(methodType, new LambdaForm("collect", invokerType.parameterCount(), arguments));
    }
    
    @LambdaForm.Hidden
    static MethodHandle selectAlternative(boolean b, MethodHandle methodHandle, MethodHandle methodHandle2) {
        if (b) {
            return methodHandle;
        }
        return methodHandle2;
    }
    
    @LambdaForm.Hidden
    static boolean profileBoolean(boolean b, int[] array) {
        int n = b;
        try {
            array[n] = Math.addExact(array[n], 1);
        }
        catch (ArithmeticException ex) {
            array[n] /= 2;
        }
        return b;
    }
    
    static MethodHandle makeGuardWithTest(MethodHandle methodHandle, MethodHandle methodHandle2, MethodHandle methodHandle3) {
        MethodType type = methodHandle2.type();
        assert methodHandle.type().equals(type.changeReturnType(Boolean.TYPE)) && methodHandle3.type().equals(type);
        LambdaForm guardWithTestForm = makeGuardWithTestForm(type.basicType());
        BoundMethodHandle boundMethodHandle;
        try {
            if (MethodHandleStatics.PROFILE_GWT) {
                boundMethodHandle = BoundMethodHandle.speciesData_LLLL().constructor().invokeBasic(type, guardWithTestForm, methodHandle, profile(methodHandle2), profile(methodHandle3), new int[2]);
            }
            else {
                boundMethodHandle = BoundMethodHandle.speciesData_LLL().constructor().invokeBasic(type, guardWithTestForm, methodHandle, profile(methodHandle2), profile(methodHandle3));
            }
        }
        catch (Throwable t) {
            throw MethodHandleStatics.uncaughtException(t);
        }
        assert boundMethodHandle.type() == type;
        return boundMethodHandle;
    }
    
    static MethodHandle profile(MethodHandle methodHandle) {
        if (MethodHandleStatics.DONT_INLINE_THRESHOLD >= 0) {
            return makeBlockInlningWrapper(methodHandle);
        }
        return methodHandle;
    }
    
    static MethodHandle makeBlockInlningWrapper(MethodHandle methodHandle) {
        return new MethodHandleImpl.CountingWrapper(methodHandle, (LambdaForm)MethodHandleImpl.PRODUCE_BLOCK_INLINING_FORM.apply(methodHandle), MethodHandleImpl.PRODUCE_BLOCK_INLINING_FORM, MethodHandleImpl.PRODUCE_REINVOKER_FORM, MethodHandleStatics.DONT_INLINE_THRESHOLD);
    }
    
    static LambdaForm makeGuardWithTestForm(MethodType methodType) {
        LambdaForm cachedLambdaForm = methodType.form().cachedLambdaForm(17);
        if (cachedLambdaForm != null) {
            return cachedLambdaForm;
        }
        int to = 1 + methodType.parameterCount();
        int n = to++;
        int n2 = to++;
        int n3 = to++;
        int n4 = MethodHandleStatics.PROFILE_GWT ? to++ : -1;
        int n5 = to++;
        int n6 = (n4 != -1) ? to++ : -1;
        int n7 = to - 1;
        int n8 = to++;
        int n9 = to++;
        assert n9 == n8 + 1;
        MethodType invokerType = methodType.invokerType();
        LambdaForm.Name[] arguments = LambdaForm.arguments(to - to, invokerType);
        BoundMethodHandle.SpeciesData speciesData = (n4 != -1) ? BoundMethodHandle.speciesData_LLLL() : BoundMethodHandle.speciesData_LLL();
        arguments[0] = arguments[0].withConstraint(speciesData);
        arguments[n] = new LambdaForm.Name(speciesData.getterFunction(0), new Object[] { arguments[0] });
        arguments[n2] = new LambdaForm.Name(speciesData.getterFunction(1), new Object[] { arguments[0] });
        arguments[n3] = new LambdaForm.Name(speciesData.getterFunction(2), new Object[] { arguments[0] });
        if (n4 != -1) {
            arguments[n4] = new LambdaForm.Name(speciesData.getterFunction(3), new Object[] { arguments[0] });
        }
        Object[] copyOfRange = Arrays.copyOfRange(arguments, 0, to, Object[].class);
        MethodType basicType = methodType.changeReturnType(Boolean.TYPE).basicType();
        copyOfRange[0] = arguments[n];
        arguments[n5] = new LambdaForm.Name(basicType, copyOfRange);
        if (n6 != -1) {
            arguments[n6] = new LambdaForm.Name(MethodHandleImpl.Lazy.NF_profileBoolean, new Object[] { arguments[n5], arguments[n4] });
        }
        copyOfRange[0] = (arguments[n8] = new LambdaForm.Name(MethodHandleImpl.Lazy.MH_selectAlternative, new Object[] { arguments[n7], arguments[n2], arguments[n3] }));
        arguments[n9] = new LambdaForm.Name(methodType, copyOfRange);
        return methodType.form().setCachedLambdaForm(17, new LambdaForm("guard", invokerType.parameterCount(), arguments, true));
    }
    
    private static LambdaForm makeGuardWithCatchForm(MethodType methodType) {
        MethodType invokerType = methodType.invokerType();
        LambdaForm cachedLambdaForm = methodType.form().cachedLambdaForm(16);
        if (cachedLambdaForm != null) {
            return cachedLambdaForm;
        }
        int n = 1 + methodType.parameterCount();
        int n2 = n++;
        int n3 = n++;
        int n4 = n++;
        int n5 = n++;
        int n6 = n++;
        int n7 = n++;
        int n8 = n++;
        int n9 = n++;
        LambdaForm.Name[] arguments = LambdaForm.arguments(n - n, invokerType);
        BoundMethodHandle.SpeciesData speciesData_LLLLL = BoundMethodHandle.speciesData_LLLLL();
        arguments[0] = arguments[0].withConstraint(speciesData_LLLLL);
        arguments[n2] = new LambdaForm.Name(speciesData_LLLLL.getterFunction(0), new Object[] { arguments[0] });
        arguments[n3] = new LambdaForm.Name(speciesData_LLLLL.getterFunction(1), new Object[] { arguments[0] });
        arguments[n4] = new LambdaForm.Name(speciesData_LLLLL.getterFunction(2), new Object[] { arguments[0] });
        arguments[n5] = new LambdaForm.Name(speciesData_LLLLL.getterFunction(3), new Object[] { arguments[0] });
        arguments[n6] = new LambdaForm.Name(speciesData_LLLLL.getterFunction(4), new Object[] { arguments[0] });
        MethodHandle basicInvoker = MethodHandles.basicInvoker(methodType.changeReturnType(Object.class));
        Object[] array = new Object[basicInvoker.type().parameterCount()];
        array[0] = arguments[n5];
        System.arraycopy(arguments, 1, array, 1, n - 1);
        arguments[n7] = new LambdaForm.Name(makeIntrinsic(basicInvoker, MethodHandleImpl.Intrinsic.GUARD_WITH_CATCH), array);
        arguments[n8] = new LambdaForm.Name(MethodHandleImpl.Lazy.NF_guardWithCatch, new Object[] { arguments[n2], arguments[n3], arguments[n4], arguments[n7] });
        arguments[n9] = new LambdaForm.Name(MethodHandles.basicInvoker(MethodType.methodType(methodType.rtype(), Object.class)), new Object[] { arguments[n6], arguments[n8] });
        return methodType.form().setCachedLambdaForm(16, new LambdaForm("guardWithCatch", invokerType.parameterCount(), arguments));
    }
    
    static MethodHandle makeGuardWithCatch(MethodHandle methodHandle, Class<? extends Throwable> clazz, MethodHandle methodHandle2) {
        MethodType type = methodHandle.type();
        LambdaForm guardWithCatchForm = makeGuardWithCatchForm(type.basicType());
        MethodHandle type2 = varargsArray(type.parameterCount()).asType(type.changeReturnType(Object[].class));
        Class<Object> returnType = type.returnType();
        MethodHandle methodHandle3;
        if (returnType.isPrimitive()) {
            if (returnType == Void.TYPE) {
                methodHandle3 = ValueConversions.ignore();
            }
            else {
                methodHandle3 = ValueConversions.unboxExact(Wrapper.forPrimitiveType(type.returnType()));
            }
        }
        else {
            methodHandle3 = MethodHandles.identity(Object.class);
        }
        BoundMethodHandle.SpeciesData speciesData_LLLLL = BoundMethodHandle.speciesData_LLLLL();
        BoundMethodHandle invokeBasic;
        try {
            invokeBasic = speciesData_LLLLL.constructor().invokeBasic(type, guardWithCatchForm, methodHandle, clazz, methodHandle2, type2, methodHandle3);
        }
        catch (Throwable t) {
            throw MethodHandleStatics.uncaughtException(t);
        }
        assert invokeBasic.type() == type;
        return invokeBasic;
    }
    
    @LambdaForm.Hidden
    static Object guardWithCatch(MethodHandle methodHandle, Class<? extends Throwable> clazz, MethodHandle methodHandle2, Object... array) throws Throwable {
        try {
            return methodHandle.asFixedArity().invokeWithArguments(array);
        }
        catch (Throwable t) {
            if (!clazz.isInstance(t)) {
                throw t;
            }
            return methodHandle2.asFixedArity().invokeWithArguments(prepend(t, array));
        }
    }
    
    @LambdaForm.Hidden
    private static Object[] prepend(Object o, Object[] array) {
        Object[] array2 = new Object[array.length + 1];
        array2[0] = o;
        System.arraycopy(array, 0, array2, 1, array.length);
        return array2;
    }
    
    static MethodHandle throwException(MethodType methodType) {
        assert !(!Throwable.class.isAssignableFrom(methodType.parameterType(0)));
        int parameterCount = methodType.parameterCount();
        if (parameterCount > 1) {
            return MethodHandles.dropArguments(throwException(methodType.dropParameterTypes(1, parameterCount)), 1, methodType.parameterList().subList(1, parameterCount));
        }
        return makePairwiseConvert(MethodHandleImpl.Lazy.NF_throwException.resolvedHandle(), methodType, false, true);
    }
    
    static <T extends Throwable> Empty throwException(T t) throws T, Throwable {
        throw t;
    }
    
    static MethodHandle fakeMethodHandleInvoke(MemberName memberName) {
        assert !(!memberName.isMethodHandleInvoke());
        String name = memberName.getName();
        int n = -1;
        switch (name.hashCode()) {
            case -1183693704: {
                if (name.equals("invoke")) {
                    n = 0;
                    break;
                }
                break;
            }
            case 941760871: {
                if (name.equals("invokeExact")) {
                    n = 1;
                    break;
                }
                break;
            }
        }
        int n2;
        switch (n) {
            case 0: {
                n2 = 0;
                break;
            }
            case 1: {
                n2 = 1;
                break;
            }
            default: {
                throw new InternalError(memberName.getName());
            }
        }
        MethodHandle methodHandle = MethodHandleImpl.FAKE_METHOD_HANDLE_INVOKE[n2];
        if (methodHandle != null) {
            return methodHandle;
        }
        MethodHandle bindTo = throwException(MethodType.methodType(Object.class, UnsupportedOperationException.class, new Class[] { MethodHandle.class, Object[].class })).bindTo(new UnsupportedOperationException("cannot reflectively invoke MethodHandle"));
        if (!memberName.getInvocationType().equals(bindTo.type())) {
            throw new InternalError(memberName.toString());
        }
        MethodHandle varargsCollector = bindTo.withInternalMemberName(memberName, false).asVarargsCollector(Object[].class);
        assert !(!memberName.isVarargs());
        return MethodHandleImpl.FAKE_METHOD_HANDLE_INVOKE[n2] = varargsCollector;
    }
    
    static MethodHandle bindCaller(MethodHandle methodHandle, Class<?> clazz) {
        return MethodHandleImpl.BindCaller.bindCaller(methodHandle, clazz);
    }
    
    static MethodHandle makeWrappedMember(MethodHandle methodHandle, MemberName memberName, boolean b) {
        if (memberName.equals(methodHandle.internalMemberName()) && b == methodHandle.isInvokeSpecial()) {
            return methodHandle;
        }
        return new MethodHandleImpl.WrappedMember(methodHandle, methodHandle.type(), memberName, b);
    }
    
    static MethodHandle makeIntrinsic(MethodHandle methodHandle, MethodHandleImpl.Intrinsic intrinsic) {
        if (intrinsic == methodHandle.intrinsicName()) {
            return methodHandle;
        }
        return new MethodHandleImpl.IntrinsicMethodHandle(methodHandle, intrinsic);
    }
    
    static MethodHandle makeIntrinsic(MethodType methodType, LambdaForm lambdaForm, MethodHandleImpl.Intrinsic intrinsic) {
        return new MethodHandleImpl.IntrinsicMethodHandle(SimpleMethodHandle.make(methodType, lambdaForm), intrinsic);
    }
    
    private static MethodHandle findCollector(String s, int n, Class<?> clazz, Class<?>... array) {
        MethodType insertParameterTypes = MethodType.genericMethodType(n).changeReturnType(clazz).insertParameterTypes(0, array);
        try {
            return MethodHandles.Lookup.IMPL_LOOKUP.findStatic(MethodHandleImpl.class, s, insertParameterTypes);
        }
        catch (ReflectiveOperationException ex) {
            return null;
        }
    }
    
    private static Object[] makeArray(Object... array) {
        return array;
    }
    
    private static Object[] array() {
        return MethodHandleImpl.NO_ARGS_ARRAY;
    }
    
    private static Object[] array(Object o) {
        return makeArray(new Object[] { o });
    }
    
    private static Object[] array(Object o, Object o2) {
        return makeArray(new Object[] { o, o2 });
    }
    
    private static Object[] array(Object o, Object o2, Object o3) {
        return makeArray(new Object[] { o, o2, o3 });
    }
    
    private static Object[] array(Object o, Object o2, Object o3, Object o4) {
        return makeArray(new Object[] { o, o2, o3, o4 });
    }
    
    private static Object[] array(Object o, Object o2, Object o3, Object o4, Object o5) {
        return makeArray(new Object[] { o, o2, o3, o4, o5 });
    }
    
    private static Object[] array(Object o, Object o2, Object o3, Object o4, Object o5, Object o6) {
        return makeArray(new Object[] { o, o2, o3, o4, o5, o6 });
    }
    
    private static Object[] array(Object o, Object o2, Object o3, Object o4, Object o5, Object o6, Object o7) {
        return makeArray(new Object[] { o, o2, o3, o4, o5, o6, o7 });
    }
    
    private static Object[] array(Object o, Object o2, Object o3, Object o4, Object o5, Object o6, Object o7, Object o8) {
        return makeArray(new Object[] { o, o2, o3, o4, o5, o6, o7, o8 });
    }
    
    private static Object[] array(Object o, Object o2, Object o3, Object o4, Object o5, Object o6, Object o7, Object o8, Object o9) {
        return makeArray(new Object[] { o, o2, o3, o4, o5, o6, o7, o8, o9 });
    }
    
    private static Object[] array(Object o, Object o2, Object o3, Object o4, Object o5, Object o6, Object o7, Object o8, Object o9, Object o10) {
        return makeArray(new Object[] { o, o2, o3, o4, o5, o6, o7, o8, o9, o10 });
    }
    
    private static MethodHandle[] makeArrays() {
        ArrayList<MethodHandle> list = new ArrayList();
        while (true) {
            MethodHandle collector = findCollector("array", list.size(), Object[].class, new Class[0]);
            if (collector == null) {
                break;
            }
            list.add(makeIntrinsic(collector, MethodHandleImpl.Intrinsic.NEW_ARRAY));
        }
        assert list.size() == 11;
        return (MethodHandle[])list.toArray(new MethodHandle[MethodHandleImpl.MAX_ARITY + 1]);
    }
    
    private static Object[] fillNewArray(Integer n, Object[] array) {
        Object[] array2 = new Object[(int)n];
        fillWithArguments(array2, 0, array);
        return array2;
    }
    
    private static Object[] fillNewTypedArray(Object[] original, Integer n, Object[] array) {
        Object[] copy = Arrays.copyOf(original, (int)n);
        assert copy.getClass() != Object[].class;
        fillWithArguments(copy, 0, array);
        return copy;
    }
    
    private static void fillWithArguments(Object[] array, int n, Object... array2) {
        System.arraycopy(array2, 0, array, n, array2.length);
    }
    
    private static Object[] fillArray(Integer n, Object[] array, Object o) {
        fillWithArguments(array, (int)n, new Object[] { o });
        return array;
    }
    
    private static Object[] fillArray(Integer n, Object[] array, Object o, Object o2) {
        fillWithArguments(array, (int)n, new Object[] { o, o2 });
        return array;
    }
    
    private static Object[] fillArray(Integer n, Object[] array, Object o, Object o2, Object o3) {
        fillWithArguments(array, (int)n, new Object[] { o, o2, o3 });
        return array;
    }
    
    private static Object[] fillArray(Integer n, Object[] array, Object o, Object o2, Object o3, Object o4) {
        fillWithArguments(array, (int)n, new Object[] { o, o2, o3, o4 });
        return array;
    }
    
    private static Object[] fillArray(Integer n, Object[] array, Object o, Object o2, Object o3, Object o4, Object o5) {
        fillWithArguments(array, (int)n, new Object[] { o, o2, o3, o4, o5 });
        return array;
    }
    
    private static Object[] fillArray(Integer n, Object[] array, Object o, Object o2, Object o3, Object o4, Object o5, Object o6) {
        fillWithArguments(array, (int)n, new Object[] { o, o2, o3, o4, o5, o6 });
        return array;
    }
    
    private static Object[] fillArray(Integer n, Object[] array, Object o, Object o2, Object o3, Object o4, Object o5, Object o6, Object o7) {
        fillWithArguments(array, (int)n, new Object[] { o, o2, o3, o4, o5, o6, o7 });
        return array;
    }
    
    private static Object[] fillArray(Integer n, Object[] array, Object o, Object o2, Object o3, Object o4, Object o5, Object o6, Object o7, Object o8) {
        fillWithArguments(array, (int)n, new Object[] { o, o2, o3, o4, o5, o6, o7, o8 });
        return array;
    }
    
    private static Object[] fillArray(Integer n, Object[] array, Object o, Object o2, Object o3, Object o4, Object o5, Object o6, Object o7, Object o8, Object o9) {
        fillWithArguments(array, (int)n, new Object[] { o, o2, o3, o4, o5, o6, o7, o8, o9 });
        return array;
    }
    
    private static Object[] fillArray(Integer n, Object[] array, Object o, Object o2, Object o3, Object o4, Object o5, Object o6, Object o7, Object o8, Object o9, Object o10) {
        fillWithArguments(array, (int)n, new Object[] { o, o2, o3, o4, o5, o6, o7, o8, o9, o10 });
        return array;
    }
    
    private static MethodHandle[] makeFillArrays() {
        ArrayList<MethodHandle> list = new ArrayList();
        list.add(null);
        while (true) {
            MethodHandle collector = findCollector("fillArray", list.size(), Object[].class, new Class[] { Integer.class, Object[].class });
            if (collector == null) {
                break;
            }
            list.add(collector);
        }
        assert list.size() == 11;
        return (MethodHandle[])list.toArray(new MethodHandle[0]);
    }
    
    private static Object copyAsPrimitiveArray(Wrapper wrapper, Object... values) {
        Object array = wrapper.makeArray(values.length);
        wrapper.copyArrayUnboxing(values, 0, array, 0, values.length);
        return array;
    }
    
    static MethodHandle varargsArray(int n) {
        MethodHandle methodHandle = MethodHandleImpl.Lazy.ARRAYS[n];
        if (methodHandle != null) {
            return methodHandle;
        }
        MethodHandle methodHandle2 = findCollector("array", n, Object[].class, new Class[0]);
        if (methodHandle2 != null) {
            methodHandle2 = makeIntrinsic(methodHandle2, MethodHandleImpl.Intrinsic.NEW_ARRAY);
        }
        if (methodHandle2 != null) {
            return MethodHandleImpl.Lazy.ARRAYS[n] = methodHandle2;
        }
        MethodHandle buildVarargsArray = buildVarargsArray(MethodHandleImpl.Lazy.MH_fillNewArray, MethodHandleImpl.Lazy.MH_arrayIdentity, n);
        assert !(!assertCorrectArity(buildVarargsArray, n));
        return MethodHandleImpl.Lazy.ARRAYS[n] = makeIntrinsic(buildVarargsArray, MethodHandleImpl.Intrinsic.NEW_ARRAY);
    }
    
    private static boolean assertCorrectArity(MethodHandle obj, int i) {
        assert obj.type().parameterCount() == i : "arity != " + i + ": " + obj;
        return true;
    }
    
    static <T> T[] identity(T[] array) {
        return array;
    }
    
    private static MethodHandle buildVarargsArray(MethodHandle methodHandle, MethodHandle collectArguments, int n) {
        int min = Math.min(n, 10);
        int n2 = n - min;
        MethodHandle methodHandle2 = methodHandle.bindTo(Integer.valueOf(n)).asCollector(Object[].class, min);
        if (n2 > 0) {
            MethodHandle fillToRight = fillToRight(10 + n2);
            if (collectArguments == MethodHandleImpl.Lazy.MH_arrayIdentity) {
                collectArguments = fillToRight;
            }
            else {
                collectArguments = MethodHandles.collectArguments(collectArguments, 0, fillToRight);
            }
        }
        if (collectArguments == MethodHandleImpl.Lazy.MH_arrayIdentity) {
            return methodHandle2;
        }
        methodHandle2 = MethodHandles.collectArguments(collectArguments, 0, methodHandle2);
        return methodHandle2;
    }
    
    private static MethodHandle fillToRight(int n) {
        MethodHandle methodHandle = MethodHandleImpl.FILL_ARRAY_TO_RIGHT[n];
        if (methodHandle != null) {
            return methodHandle;
        }
        MethodHandle buildFiller = buildFiller(n);
        assert !(!assertCorrectArity(buildFiller, n - 10 + 1));
        return MethodHandleImpl.FILL_ARRAY_TO_RIGHT[n] = buildFiller;
    }
    
    private static MethodHandle buildFiller(int n) {
        if (n <= 10) {
            return MethodHandleImpl.Lazy.MH_arrayIdentity;
        }
        int n2 = n % 10;
        int i = n - n2;
        if (n2 == 0) {
            i = n - (n2 = 10);
            if (MethodHandleImpl.FILL_ARRAY_TO_RIGHT[i] == null) {
                for (int j = 0; j < i; j += 10) {
                    if (j > 10) {
                        fillToRight(j);
                    }
                }
            }
        }
        if (i < 10) {
            n2 = n - (i = 10);
        }
        assert n2 > 0;
        MethodHandle fillToRight = fillToRight(i);
        MethodHandle bindTo = MethodHandleImpl.Lazy.FILL_ARRAYS[n2].bindTo(Integer.valueOf(i));
        assert fillToRight.type().parameterCount() == 1 + i - 10;
        assert bindTo.type().parameterCount() == 1 + n2;
        if (i == 10) {
            return bindTo;
        }
        return MethodHandles.collectArguments(bindTo, 0, fillToRight);
    }
    
    static MethodHandle varargsArray(Class<?> obj, int n) {
        Class<Object> componentType = obj.getComponentType();
        if (componentType == null) {
            throw new IllegalArgumentException("not an array: " + obj);
        }
        if (n >= 126) {
            if (n <= 254 && componentType.isPrimitive()) {
                n *= Wrapper.forPrimitiveType(componentType).stackSlots();
            }
            if (n > 254) {
                throw new IllegalArgumentException("too many arguments: " + obj.getSimpleName() + ", length " + n);
            }
        }
        if (componentType == Object.class) {
            return varargsArray(n);
        }
        MethodHandle[] array = (MethodHandle[])MethodHandleImpl.TYPED_COLLECTORS.get(componentType);
        MethodHandle methodHandle = (n < array.length) ? array[n] : null;
        if (methodHandle != null) {
            return methodHandle;
        }
        MethodHandle methodHandle2;
        if (n == 0) {
            methodHandle2 = MethodHandles.constant(obj, Array.newInstance(obj.getComponentType(), 0));
        }
        else if (componentType.isPrimitive()) {
            methodHandle2 = buildVarargsArray(MethodHandleImpl.Lazy.MH_fillNewArray, buildArrayProducer(obj), n);
        }
        else {
            methodHandle2 = buildVarargsArray(MethodHandleImpl.Lazy.MH_fillNewTypedArray.bindTo(Arrays.copyOf(MethodHandleImpl.NO_ARGS_ARRAY, 0, obj.asSubclass(Object[].class))), MethodHandleImpl.Lazy.MH_arrayIdentity, n);
        }
        MethodHandle intrinsic = makeIntrinsic(methodHandle2.asType(MethodType.methodType(obj, Collections.nCopies(n, componentType))), MethodHandleImpl.Intrinsic.NEW_ARRAY);
        assert !(!assertCorrectArity(intrinsic, n));
        if (n < array.length) {
            array[n] = intrinsic;
            return intrinsic;
        }
        return intrinsic;
    }
    
    private static MethodHandle buildArrayProducer(Class<?> clazz) {
        Class<Object> componentType = clazz.getComponentType();
        assert !(!componentType.isPrimitive());
        return MethodHandleImpl.Lazy.MH_copyAsPrimitiveArray.bindTo(Wrapper.forPrimitiveType(componentType));
    }
    
    static void assertSame(Object o, Object o2) {
        if (o != o2) {
            throw MethodHandleStatics.newInternalError(String.format("mh1 != mh2: mh1 = %s (form: %s); mh2 = %s (form: %s)", new Object[] { o, ((MethodHandle)o).form, o2, ((MethodHandle)o2).form }));
        }
        return;
    }
    
    static MethodHandle[] access$000() {
        return makeArrays();
    }
    
    static MethodHandle[] access$100() {
        return makeFillArrays();
    }
    
    static {
        final Object[] array = { Integer.valueOf(255) };
        AccessController.doPrivileged(new PrivilegedAction<Void>() {
            @Override
            public Void run() {
                array[0] = Integer.getInteger(MethodHandleImpl.class.getName() + ".MAX_ARITY", 255);
                return null;
            }
        });
        MAX_ARITY = (int)array[0];
        PRODUCE_BLOCK_INLINING_FORM = new Function<MethodHandle, LambdaForm>() {
            @Override
            public LambdaForm apply(MethodHandle methodHandle) {
                return DelegatingMethodHandle.makeReinvokerForm(methodHandle, 9, MethodHandleImpl.CountingWrapper.class, "reinvoker.dontInline", false, DelegatingMethodHandle.NF_getTarget, MethodHandleImpl.CountingWrapper.NF_maybeStopCounting);
            }
        };
        PRODUCE_REINVOKER_FORM = new Function<MethodHandle, LambdaForm>() {
            @Override
            public LambdaForm apply(MethodHandle methodHandle) {
                return DelegatingMethodHandle.makeReinvokerForm(methodHandle, 8, DelegatingMethodHandle.class, DelegatingMethodHandle.NF_getTarget);
            }
        };
        MethodHandleImpl.FAKE_METHOD_HANDLE_INVOKE = new MethodHandle[2];
        NO_ARGS_ARRAY = new Object[0];
        FILL_ARRAY_TO_RIGHT = new MethodHandle[MethodHandleImpl.MAX_ARITY + 1];
        TYPED_COLLECTORS = new ClassValue<MethodHandle[]>() {
            @Override
            protected MethodHandle[] computeValue(Class<?> clazz) {
                return new MethodHandle[256];
            }
        };
    }
    
    static final class ArrayAccessor
    {
        static final int GETTER_INDEX = 0;
        static final int SETTER_INDEX = 1;
        static final int INDEX_LIMIT = 2;
        static final ClassValue<MethodHandle[]> TYPED_ACCESSORS;
        static final MethodHandle OBJECT_ARRAY_GETTER;
        static final MethodHandle OBJECT_ARRAY_SETTER;
        
        static int getElementI(int[] array, int n) {
            return array[n];
        }
        
        static long getElementJ(long[] array, int n) {
            return array[n];
        }
        
        static float getElementF(float[] array, int n) {
            return array[n];
        }
        
        static double getElementD(double[] array, int n) {
            return array[n];
        }
        
        static boolean getElementZ(boolean[] array, int n) {
            return array[n];
        }
        
        static byte getElementB(byte[] array, int n) {
            return array[n];
        }
        
        static short getElementS(short[] array, int n) {
            return array[n];
        }
        
        static char getElementC(char[] array, int n) {
            return array[n];
        }
        
        static Object getElementL(Object[] array, int n) {
            return array[n];
        }
        
        static void setElementI(int[] array, int n, int n2) {
            array[n] = n2;
        }
        
        static void setElementJ(long[] array, int n, long n2) {
            array[n] = n2;
        }
        
        static void setElementF(float[] array, int n, float n2) {
            array[n] = n2;
        }
        
        static void setElementD(double[] array, int n, double n2) {
            array[n] = n2;
        }
        
        static void setElementZ(boolean[] array, int n, boolean b) {
            array[n] = b;
        }
        
        static void setElementB(byte[] array, int n, byte b) {
            array[n] = b;
        }
        
        static void setElementS(short[] array, int n, short n2) {
            array[n] = n2;
        }
        
        static void setElementC(char[] array, int n, char c) {
            array[n] = c;
        }
        
        static void setElementL(Object[] array, int n, Object o) {
            array[n] = o;
        }
        
        static String name(Class<?> clazz, boolean b) {
            Class<Object> componentType = clazz.getComponentType();
            if (componentType == null) {
                throw MethodHandleStatics.newIllegalArgumentException("not an array", clazz);
            }
            return (b ? "setElement" : "getElement") + Wrapper.basicTypeChar(componentType);
        }
        
        static MethodType type(Class<?> clazz, boolean b) {
            Class<Object> componentType = clazz.getComponentType();
            if (!componentType.isPrimitive()) {
                clazz = Object[].class;
                componentType = Object.class;
            }
            return b ? MethodType.methodType(Void.TYPE, clazz, new Class[] { Integer.TYPE, componentType }) : MethodType.methodType(componentType, clazz, new Class[] { Integer.TYPE });
        }
        
        static MethodType correctType(Class<?> clazz, boolean b) {
            Class<Object> componentType = clazz.getComponentType();
            return b ? MethodType.methodType(Void.TYPE, clazz, new Class[] { Integer.TYPE, componentType }) : MethodType.methodType(componentType, clazz, new Class[] { Integer.TYPE });
        }
        
        static MethodHandle getAccessor(Class<?> clazz, boolean b) {
            String name = name(clazz, b);
            MethodType type = type(clazz, b);
            try {
                return MethodHandles.Lookup.IMPL_LOOKUP.findStatic(MethodHandleImpl.ArrayAccessor.class, name, type);
            }
            catch (ReflectiveOperationException ex) {
                throw MethodHandleStatics.uncaughtException(ex);
            }
        }
        
        static {
            TYPED_ACCESSORS = new ClassValue<MethodHandle[]>() {
                @Override
                protected MethodHandle[] computeValue(Class<?> clazz) {
                    return new MethodHandle[2];
                }
            };
            MethodHandle[] array = (MethodHandle[])MethodHandleImpl.ArrayAccessor.TYPED_ACCESSORS.get(Object[].class);
            array[0] = (OBJECT_ARRAY_GETTER = MethodHandleImpl.makeIntrinsic(getAccessor(Object[].class, false), MethodHandleImpl.Intrinsic.ARRAY_LOAD));
            array[1] = (OBJECT_ARRAY_SETTER = MethodHandleImpl.makeIntrinsic(getAccessor(Object[].class, true), MethodHandleImpl.Intrinsic.ARRAY_STORE));
            assert !(!InvokerBytecodeGenerator.isStaticallyInvocable(MethodHandleImpl.ArrayAccessor.OBJECT_ARRAY_GETTER.internalMemberName()));
            assert !(!InvokerBytecodeGenerator.isStaticallyInvocable(MethodHandleImpl.ArrayAccessor.OBJECT_ARRAY_SETTER.internalMemberName()));
            return;
        }
    }
    
    private static final class AsVarargsCollector extends DelegatingMethodHandle
    {
        private final MethodHandle target;
        private final Class<?> arrayType;
        @Stable
        private MethodHandle asCollectorCache;
        
        AsVarargsCollector(MethodHandle methodHandle, Class<?> clazz) {
            this(methodHandle.type(), methodHandle, clazz);
        }
        
        AsVarargsCollector(MethodType methodType, MethodHandle target, Class<?> arrayType) {
            super(methodType, target);
            this.target = target;
            this.arrayType = arrayType;
            this.asCollectorCache = target.asCollector(arrayType, 0);
        }
        
        @Override
        public boolean isVarargsCollector() {
            return true;
        }
        
        @Override
        protected MethodHandle getTarget() {
            return this.target;
        }
        
        @Override
        public MethodHandle asFixedArity() {
            return this.target;
        }
        
        @Override
        MethodHandle setVarargs(MemberName memberName) {
            if (memberName.isVarargs()) {
                return this;
            }
            return this.asFixedArity();
        }
        
        public MethodHandle asTypeUncached(MethodType methodType) {
            MethodType type = this.type();
            int n = type.parameterCount() - 1;
            int parameterCount = methodType.parameterCount();
            if (parameterCount == n + 1 && type.parameterType(n).isAssignableFrom(methodType.parameterType(n))) {
                return this.asTypeCache = this.asFixedArity().asType(methodType);
            }
            MethodHandle asCollectorCache = this.asCollectorCache;
            if (asCollectorCache != null && asCollectorCache.type().parameterCount() == parameterCount) {
                return this.asTypeCache = asCollectorCache.asType(methodType);
            }
            int n2 = parameterCount - n;
            MethodHandle collector;
            try {
                collector = this.asFixedArity().asCollector(this.arrayType, n2);
                assert collector.type().parameterCount() == parameterCount : "newArity=" + parameterCount + " but collector=" + collector;
            }
            catch (IllegalArgumentException ex) {
                throw new WrongMethodTypeException("cannot build collector", ex);
            }
            this.asCollectorCache = collector;
            return this.asTypeCache = collector.asType(methodType);
        }
        
        @Override
        boolean viewAsTypeChecks(MethodType methodType, boolean b) {
            super.viewAsTypeChecks(methodType, true);
            if (b) {
                return true;
            }
            assert !(!this.type().lastParameterType().getComponentType().isAssignableFrom(methodType.lastParameterType().getComponentType())) : Arrays.asList(new Object[] { this, methodType });
            return true;
        }
    }
    
    static class Lazy
    {
        private static final Class<?> MHI;
        private static final MethodHandle[] ARRAYS;
        private static final MethodHandle[] FILL_ARRAYS;
        static final LambdaForm.NamedFunction NF_checkSpreadArgument;
        static final LambdaForm.NamedFunction NF_guardWithCatch;
        static final LambdaForm.NamedFunction NF_throwException;
        static final LambdaForm.NamedFunction NF_profileBoolean;
        static final MethodHandle MH_castReference;
        static final MethodHandle MH_selectAlternative;
        static final MethodHandle MH_copyAsPrimitiveArray;
        static final MethodHandle MH_fillNewTypedArray;
        static final MethodHandle MH_fillNewArray;
        static final MethodHandle MH_arrayIdentity;
        
        static {
            MHI = MethodHandleImpl.class;
            ARRAYS = MethodHandleImpl.access$000();
            FILL_ARRAYS = MethodHandleImpl.access$100();
            try {
                NF_checkSpreadArgument = new LambdaForm.NamedFunction(MethodHandleImpl.Lazy.MHI.getDeclaredMethod("checkSpreadArgument", new Class[] { Object.class, Integer.TYPE }));
                NF_guardWithCatch = new LambdaForm.NamedFunction(MethodHandleImpl.Lazy.MHI.getDeclaredMethod("guardWithCatch", new Class[] { MethodHandle.class, Class.class, MethodHandle.class, Object[].class }));
                NF_throwException = new LambdaForm.NamedFunction(MethodHandleImpl.Lazy.MHI.getDeclaredMethod("throwException", new Class[] { Throwable.class }));
                NF_profileBoolean = new LambdaForm.NamedFunction(MethodHandleImpl.Lazy.MHI.getDeclaredMethod("profileBoolean", new Class[] { Boolean.TYPE, int[].class }));
                MethodHandleImpl.Lazy.NF_checkSpreadArgument.resolve();
                MethodHandleImpl.Lazy.NF_guardWithCatch.resolve();
                MethodHandleImpl.Lazy.NF_throwException.resolve();
                MethodHandleImpl.Lazy.NF_profileBoolean.resolve();
                MH_castReference = MethodHandles.Lookup.IMPL_LOOKUP.findStatic(MethodHandleImpl.Lazy.MHI, "castReference", MethodType.methodType(Object.class, Class.class, new Class[] { Object.class }));
                MH_copyAsPrimitiveArray = MethodHandles.Lookup.IMPL_LOOKUP.findStatic(MethodHandleImpl.Lazy.MHI, "copyAsPrimitiveArray", MethodType.methodType(Object.class, Wrapper.class, new Class[] { Object[].class }));
                MH_arrayIdentity = MethodHandles.Lookup.IMPL_LOOKUP.findStatic(MethodHandleImpl.Lazy.MHI, "identity", MethodType.methodType(Object[].class, Object[].class));
                MH_fillNewArray = MethodHandles.Lookup.IMPL_LOOKUP.findStatic(MethodHandleImpl.Lazy.MHI, "fillNewArray", MethodType.methodType(Object[].class, Integer.class, new Class[] { Object[].class }));
                MH_fillNewTypedArray = MethodHandles.Lookup.IMPL_LOOKUP.findStatic(MethodHandleImpl.Lazy.MHI, "fillNewTypedArray", MethodType.methodType(Object[].class, Object[].class, new Class[] { Integer.class, Object[].class }));
                MH_selectAlternative = MethodHandleImpl.makeIntrinsic(MethodHandles.Lookup.IMPL_LOOKUP.findStatic(MethodHandleImpl.Lazy.MHI, "selectAlternative", MethodType.methodType(MethodHandle.class, Boolean.TYPE, new Class[] { MethodHandle.class, MethodHandle.class })), MethodHandleImpl.Intrinsic.SELECT_ALTERNATIVE);
            }
            catch (ReflectiveOperationException ex) {
                throw MethodHandleStatics.newInternalError(ex);
            }
        }
    }
    
    static class CountingWrapper extends DelegatingMethodHandle
    {
        private final MethodHandle target;
        private int count;
        private Function<MethodHandle, LambdaForm> countingFormProducer;
        private Function<MethodHandle, LambdaForm> nonCountingFormProducer;
        private volatile boolean isCounting;
        static final LambdaForm.NamedFunction NF_maybeStopCounting;
        
        private CountingWrapper(MethodHandle target, LambdaForm lambdaForm, Function<MethodHandle, LambdaForm> countingFormProducer, Function<MethodHandle, LambdaForm> nonCountingFormProducer, int count) {
            super(target.type(), lambdaForm);
            this.target = target;
            this.count = count;
            this.countingFormProducer = countingFormProducer;
            this.nonCountingFormProducer = nonCountingFormProducer;
            this.isCounting = (count > 0);
        }
        
        @LambdaForm.Hidden
        @Override
        protected MethodHandle getTarget() {
            return this.target;
        }
        
        public MethodHandle asTypeUncached(MethodType methodType) {
            MethodHandle type = this.target.asType(methodType);
            MethodHandleImpl.CountingWrapper asTypeCache;
            if (this.isCounting) {
                asTypeCache = new MethodHandleImpl.CountingWrapper(type, (LambdaForm)this.countingFormProducer.apply(type), this.countingFormProducer, this.nonCountingFormProducer, MethodHandleStatics.DONT_INLINE_THRESHOLD);
            }
            else {
                asTypeCache = type;
            }
            return this.asTypeCache = asTypeCache;
        }
        
        boolean countDown() {
            if (this.count <= 0) {
                if (this.isCounting) {
                    this.isCounting = false;
                    return true;
                }
                return false;
            }
            else {
                --this.count;
                return false;
            }
        }
        
        @LambdaForm.Hidden
        static void maybeStopCounting(Object o) {
            MethodHandleImpl.CountingWrapper countingWrapper = (MethodHandleImpl.CountingWrapper)o;
            if (countingWrapper.countDown()) {
                LambdaForm lambdaForm = (LambdaForm)countingWrapper.nonCountingFormProducer.apply(countingWrapper.target);
                lambdaForm.compileToBytecode();
                countingWrapper.updateForm(lambdaForm);
            }
        }
        
        static {
            Class<Object> clazz = MethodHandleImpl.CountingWrapper.class;
            try {
                NF_maybeStopCounting = new LambdaForm.NamedFunction(clazz.getDeclaredMethod("maybeStopCounting", new Class[] { Object.class }));
            }
            catch (ReflectiveOperationException ex) {
                throw MethodHandleStatics.newInternalError(ex);
            }
        }
    }
    
    private static class BindCaller
    {
        private static ClassValue<MethodHandle> CV_makeInjectedInvoker;
        private static final MethodHandle MH_checkCallerClass;
        private static final byte[] T_BYTES;
        
        static MethodHandle bindCaller(MethodHandle methodHandle, Class<?> type) {
            if (type == null || type.isArray() || type.isPrimitive() || type.getName().startsWith("java.") || type.getName().startsWith("sun.")) {
                throw new InternalError();
            }
            return restoreToType(((MethodHandle)MethodHandleImpl.BindCaller.CV_makeInjectedInvoker.get(type)).bindTo(prepareForInvoker(methodHandle)), methodHandle, type);
        }
        
        private static MethodHandle makeInjectedInvoker(Class<?> clazz) {
            Class<Object> defineAnonymousClass = MethodHandleStatics.UNSAFE.defineAnonymousClass(clazz, MethodHandleImpl.BindCaller.T_BYTES, null);
            if (clazz.getClassLoader() != defineAnonymousClass.getClassLoader()) {
                throw new InternalError(clazz.getName() + " (CL)");
            }
            try {
                if (clazz.getProtectionDomain() != defineAnonymousClass.getProtectionDomain()) {
                    throw new InternalError(clazz.getName() + " (PD)");
                }
            }
            catch (SecurityException ex) {}
            try {
                MethodHandles.Lookup.IMPL_LOOKUP.findStatic(defineAnonymousClass, "init", MethodType.methodType(Void.TYPE)).invokeExact();
            }
            catch (Throwable t) {
                throw MethodHandleStatics.uncaughtException(t);
            }
            MethodHandle static1;
            try {
                static1 = MethodHandles.Lookup.IMPL_LOOKUP.findStatic(defineAnonymousClass, "invoke_V", MethodType.methodType(Object.class, MethodHandle.class, new Class[] { Object[].class }));
            }
            catch (ReflectiveOperationException ex2) {
                throw MethodHandleStatics.uncaughtException(ex2);
            }
            try {
                static1.invokeExact(prepareForInvoker(MethodHandleImpl.BindCaller.MH_checkCallerClass), new Object[] { clazz, defineAnonymousClass });
            }
            catch (Throwable cause) {
                throw new InternalError(cause);
            }
            return static1;
        }
        
        private static MethodHandle prepareForInvoker(MethodHandle fixedArity) {
            fixedArity = fixedArity.asFixedArity();
            MethodType type = fixedArity.type();
            int parameterCount = type.parameterCount();
            MethodHandle type2 = fixedArity.asType(type.generic());
            type2.internalForm().compileToBytecode();
            MethodHandle spreader = type2.asSpreader(Object[].class, parameterCount);
            spreader.internalForm().compileToBytecode();
            return spreader;
        }
        
        private static MethodHandle restoreToType(MethodHandle methodHandle, MethodHandle methodHandle2, Class<?> clazz) {
            MethodType type = methodHandle2.type();
            return new MethodHandleImpl.WrappedMember(methodHandle.asCollector(Object[].class, type.parameterCount()).asType(type), type, methodHandle2.internalMemberName(), methodHandle2.isInvokeSpecial(), clazz);
        }
        
        @CallerSensitive
        private static boolean checkCallerClass(Class<?> clazz, Class<?> clazz2) {
            Class<Object> callerClass = Reflection.getCallerClass();
            if (callerClass != clazz && callerClass != clazz2) {
                throw new InternalError("found " + callerClass.getName() + ", expected " + clazz.getName() + ((clazz == clazz2) ? "" : (", or else " + clazz2.getName())));
            }
            return true;
        }
        
        static MethodHandle access$300(Class clazz) {
            return makeInjectedInvoker(clazz);
        }
        
        static {
            MethodHandleImpl.BindCaller.CV_makeInjectedInvoker = new ClassValue<MethodHandle>() {
                @Override
                protected MethodHandle computeValue(Class<?> clazz) {
                    return MethodHandleImpl.BindCaller.access$300(clazz);
                }
            };
            Class<Object> clazz = MethodHandleImpl.BindCaller.class;
            assert !(!checkCallerClass(clazz, clazz));
            try {
                MH_checkCallerClass = MethodHandles.Lookup.IMPL_LOOKUP.findStatic(clazz, "checkCallerClass", MethodType.methodType(Boolean.TYPE, Class.class, new Class[] { Class.class }));
                assert !(!MethodHandleImpl.BindCaller.MH_checkCallerClass.invokeExact(clazz, clazz));
            }
            catch (Throwable cause) {
                throw new InternalError(cause);
            }
            Object[] array = { null };
            AccessController.doPrivileged(new PrivilegedAction<Void>() {
                @Override
                public Void run() {
                    try {
                        Class<Object> clazz = MethodHandleImpl.BindCaller.T.class;
                        String name = clazz.getName();
                        String string = name.substring(name.lastIndexOf(46) + 1) + ".class";
                        URLConnection openConnection = clazz.getResource(string).openConnection();
                        int contentLength = openConnection.getContentLength();
                        byte[] b = new byte[contentLength];
                        try (final InputStream inputStream = openConnection.getInputStream()) {
                            if (inputStream.read(b) != contentLength) {
                                throw new IOException(string);
                            }
                        }
                        array[0] = b;
                    }
                    catch (IOException cause) {
                        throw new InternalError(cause);
                    }
                    return null;
                }
            });
            T_BYTES = (byte[])(byte[])array[0];
            return;
        }
        
        private static class T
        {
            static void init() {
            }
            
            static Object invoke_V(MethodHandle methodHandle, Object[] array) throws Throwable {
                return methodHandle.invokeExact(array);
            }
        }
    }
    
    private static final class WrappedMember extends DelegatingMethodHandle
    {
        private final MethodHandle target;
        private final MemberName member;
        private final Class<?> callerClass;
        private final boolean isInvokeSpecial;
        
        private WrappedMember(MethodHandle target, MethodType methodType, MemberName member, boolean isInvokeSpecial, Class<?> callerClass) {
            super(methodType, target);
            this.target = target;
            this.member = member;
            this.callerClass = callerClass;
            this.isInvokeSpecial = isInvokeSpecial;
        }
        
        @Override
        MemberName internalMemberName() {
            return this.member;
        }
        
        @Override
        Class<?> internalCallerClass() {
            return this.callerClass;
        }
        
        @Override
        boolean isInvokeSpecial() {
            return this.isInvokeSpecial;
        }
        
        @Override
        protected MethodHandle getTarget() {
            return this.target;
        }
        
        public MethodHandle asTypeUncached(MethodType methodType) {
            return this.asTypeCache = this.target.asType(methodType);
        }
    }
    
    enum Intrinsic
    {
        SELECT_ALTERNATIVE, 
        GUARD_WITH_CATCH, 
        NEW_ARRAY, 
        ARRAY_LOAD, 
        ARRAY_STORE, 
        IDENTITY, 
        ZERO, 
        NONE;
    }
    
    private static final class IntrinsicMethodHandle extends DelegatingMethodHandle
    {
        private final MethodHandle target;
        private final MethodHandleImpl.Intrinsic intrinsicName;
        
        IntrinsicMethodHandle(MethodHandle target, MethodHandleImpl.Intrinsic intrinsicName) {
            super(target.type(), target);
            this.target = target;
            this.intrinsicName = intrinsicName;
        }
        
        @Override
        protected MethodHandle getTarget() {
            return this.target;
        }
        
        @Override
        MethodHandleImpl.Intrinsic intrinsicName() {
            return this.intrinsicName;
        }
        
        public MethodHandle asTypeUncached(MethodType methodType) {
            return this.asTypeCache = this.target.asType(methodType);
        }
        
        @Override
        String internalProperties() {
            return super.internalProperties() + "\n& Intrinsic=" + this.intrinsicName;
        }
        
        @Override
        public MethodHandle asCollector(Class<?> clazz, int n) {
            if (this.intrinsicName == MethodHandleImpl.Intrinsic.IDENTITY) {
                return MethodHandleImpl.varargsArray(clazz, n).asType(this.type().asCollectorType(clazz, n));
            }
            return super.asCollector(clazz, n);
        }
    }
}
